<?php

declare(strict_types=1);

namespace mirzaev\arming_bot\models;

// Files of the project
use mirzaev\arming_bot\models\core,
	mirzaev\arming_bot\models\traits\document as arangodb_document_trait,
	mirzaev\arming_bot\models\interfaces\document as arangodb_document_interface;

// Framework for ArangoDB
use mirzaev\arangodb\collection,
	mirzaev\arangodb\document;

// Library для ArangoDB
use ArangoDBClient\Document as _document;

// Built-in libraries
use exception;

/**
 * Model of a category
 *
 * @package mirzaev\arming_bot\models
 * @author Arsen Mirzaev Tatyano-Muradovich <arsen@mirzaev.sexy>
 */
class categories extends core implements arangodb_document_interface
{
	use arangodb_document_trait;

	/**
	 * Name of the collection in ArangoDB
	 */
	public const string COLLECTION = 'THIS_COLLECTION_SHOULD_NOT_EXIST';

	/**
	 * Запись категории
	 *
	 * @param string $name Название
	 * @param arrau $labels Ярлыки для отображения в интерфейсе ['EN' => "Label"]
	 * @param array $hierarchy Иерархия вложенности (от родителей к потомкам: [ model, model ... ])
	 * @param array &$errors Registry of errors
	 *
	 * @return string|null Идентификатор (_id) документа, если создан
	 */
	public static function write(
		string $name,
		array $labels = ['RU' => 'Без названия'],
		array $hierarchy = [],
		array &$errors = []
	): string|null {
		try {
			if (collection::init(core::$arangodb->session, static::COLLECTION)) {
				// Инициализирована коллекция

				// Создание категории
				$category = document::write(
					core::$arangodb->session,
					static::COLLECTION,
					[
						'name' => $name,
						'labels' => $labels,
						'version' => ROBOT_VERSION
					]
				);

				if ($category) {
					// Создана категория

					/* if (collection::init(core::$arangodb->session, 'entry', true)) {
						// Инициализирована коллекция

						foreach ($hierarchy as $model) {
							// Перебор иерархической структуры категорий

							// Инициализация вложенной категории (следующей в массиве)
							$next = current($hierarchy);

							// Поиск ребра описывающего иерархическую связь

							document::write(core::$arangodb->session, 'entry');
						}
					} else throw new exception('Failed to initialize document collection: ' . static::COLLECTION); */
				}
			} else throw new exception('Failed to initialize edge collection: entry');
		} catch (exception $e) {
			// Write to the registry of errors
			$errors[] = [
				'text' => $e->getMessage(),
				'file' => $e->getFile(),
				'line' => $e->getLine(),
				'stack' => $e->getTrace()
			];
		}

		// Выход (провал)
		return null;
	}

	/**
	 * Поиск вхождений (подкатегории или товары)
	 *
	 * Находит вхождения через ребро entry
	 * Генерирует _type со значениями "category" и "product"
	 * относительно того есть ли у документа ещё вложения (у product вложений быть не может)
	 * Объединяет возвращаемые объекты документа с переменной _type
	 *
	 * @param string|null $category Category identifier (_id)
	 * @param string|null $filter Expression for filtering (AQL)
	 * @param string|null $sort Expression for sorting (AQL)
	 * @param int $page Страница
	 * @param int $amount Количество товаров на странице
	 * @param array &$errors Registry of errors
	 *
	 * @return array Массив с найденными вхождениями (может быть пустым)
	 */
	public static function entries(
		?string $category = null,
		?string $filter = null,
		?string $sort = 'v.promotion DESC, v.position ASC, v.created DESC',
		int $page = 1,
		int $amount = 100,
		array &$errors = []
	): array {
		try {
			if (collection::init(core::$arangodb->session, static::COLLECTION)) {
				// Инициализирована коллекция
				
				if ($documents = collection::search(
					core::$arangodb->session,
					sprintf(
						<<<AQL
						FOR v IN 1..1 INBOUND %s GRAPH hierarchy
							%s
							%s
							LIMIT %u, %u
							LET _type = (FOR v2 IN INBOUND v._id GRAPH hierarchy RETURN v2)[0] ? "category" : "product"
							RETURN MERGE(v, {_type})
						AQL,
						empty($category) ? '(FOR d IN ' . ${static::COLLECTION} . 'LIMIT 1 RETURN d._id)[0]' : "\"$category\"",
						empty($filter) ? '' : "FILTER $filter",
						empty($sort) ? '' : "SORT $sort",
						--$page <= 0 ? $page = 0 : $page * $amount,
						$amount,
					)
				)) {
					// Найдены вхождения

					// Возврат (успех)
					return is_array($documents) ? $documents : [$documents];
				} else return [];
			} else throw new exception('Failed to initialize document collection: ' . static::COLLECTION);
		} catch (exception $e) {
			// Write to the registry of errors
			$errors[] = [
				'text' => $e->getMessage(),
				'file' => $e->getFile(),
				'line' => $e->getLine(),
				'stack' => $e->getTrace()
			];
		}

		// Выход (провал)
		return [];
	}

	/**
	 * Поиск категорий и товаров
	 *
	 * Ищет категории и товары по коллекции рёбер entry из _from и _to
	 *
	 * @param array &$errors Registry of errors
	 *
	 * @return array Массив с уникализированными найденными коллекциями (может быть пустым)
	 */
	protected static function collections(
		array &$errors = []
	): array {
		try {
			if (collection::init(core::$arangodb->session, $collection = 'entry')) {
				// Инициализирована коллекция

				if ($names = collection::search(
					core::$arangodb->session,
					sprintf(
						<<<AQL
						FOR e IN %s
							COLLECT AGGREGATE 
								from_category = UNIQUE(PARSE_IDENTIFIER(e._from).collection), 
								to_category = UNIQUE(PARSE_IDENTIFIER(e._to).collection)
							RETURN UNIQUE(UNION(from_category, to_category))
						AQL,
						$collection
					)
				)) {
					// Найдены коллекции

					// Возврат (успех)
					return $names->getAll();
				} else return [];
			} else throw new exception('Failed to initialize edge collection: entry');
		} catch (exception $e) {
			// Write to the registry of errors
			$errors[] = [
				'text' => $e->getMessage(),
				'file' => $e->getFile(),
				'line' => $e->getLine(),
				'stack' => $e->getTrace()
			];
		}

		// Выход (провал)
		return [];
	}

	/**
	 * Поиск категории по названияю
	 *
	 * Перебирает все коллекции из self::collections() и ищет в них соответствие с параметром name
	 * ВНИМАНИЕ: НЕЛЬЗЯ ДОПУСКАТЬ ОДИНАКОВЫЕ НАЗВАНИЯ СРЕДИ РАЗНЫХ КАТЕГОРИЙ
	 *
	 * @param string $name Name of a collection
	 * @param array &$errors Registry of errors
	 *
	 * @return categories|null Модель имплементирующая документ с категорией, если была найдена
	 */
	public static function category(
		string $name,
		array &$errors = []
	): ?categories {
		try {
			// Инициалиация списка коллекций
			$collections = self::collections($errors);

			if (count($collections) > 0) {
				// Найдены коллекции

				// Инициализация буфера части запроса со списком коллекций для аргументов UNION()
				$union = [];
				foreach ($collections as $collection) $union[] = "FOR d IN $collection RETURN d";
				unset($collection);

				if ($document = collection::search(
					core::$arangodb->session,
					sprintf(
						<<<AQL
						FOR u IN UNION(%s)
							FILTER u.name == "%s"
							SORT u.created ASC
							LIMIT 1
							RETURN u
						AQL,
						implode(', ', $union),
						$name
					)
				)) {
					// Найдена категория

					// Инициализация названия коллекции
					$collection = explode('/', $document->getId())[0];

					if (class_exists($model = "mirzaev\\arming_bot\\models\\$collection")) {
						// Найдена модель имплементирующая документы из этой коллекции

						// Инициализация объекта модели
						$object = new $model;

						if ($object instanceof categories) {
							// Объект является инстанцией категории (то есть не товара)

							// Запись инстанции документа в объект модели
							$object->document = $document;

							// Возврат (успех)
							return $object;
						}
					}
				} else return null;
			}
		} catch (exception $e) {
			// Write to the registry of errors
			$errors[] = [
				'text' => $e->getMessage(),
				'file' => $e->getFile(),
				'line' => $e->getLine(),
				'stack' => $e->getTrace()
			];
		}

		// Выход (провал)
		return null;
	}
}
