<?php

declare(strict_types=1);

namespace mirzaev\arming_bot\models;

// Files of the project
use mirzaev\arming_bot\models\core,
	mirzaev\arming_bot\controllers\core as controller,
	mirzaev\arming_bot\models\account;

// Фреймворк Telegram
use Zanzara\Zanzara,
	Zanzara\Context,
	Zanzara\Telegram\Type\Input\InputFile,
	Zanzara\Telegram\Type\File\Document as telegram_document,
	Zanzara\Telegram\Type\File\File,
	Zanzara\Middleware\MiddlewareNode as Node;

// Library для ArangoDB
use ArangoDBClient\Document as _document;

/**
 * Model of a chat
 *
 * @package mirzaev\arming_bot\models
 * @author Arsen Mirzaev Tatyano-Muradovich <arsen@mirzaev.sexy>
 */
final class chat extends core
{
	/**
	 * Экранирование символов для Markdown
	 *
	 * @param string $text Текст для экранирования
	 * @param array $exception Символы которые будут исключены из списка для экранирования
	 *
	 * @return string Экранированный текст
	 */
	public static function unmarkdown(string $text, array $exceptions = []): string
	{
		// Инициализация реестра символом для конвертации
		$from = array_diff(
			[
				'#',
				'*',
				'_',
				'=',
				'.',
				'[',
				']',
				'(',
				')',
				'-',
				'>',
				'<',
				'!',
				'`'
			],
			$exceptions
		);

		// Инициализация реестра целей для конвертации
		$to = [];
		foreach ($from as $symbol) $to[] = "\\$symbol";

		// Конвертация и выход (успех)
		return str_replace($from, $to,	$text);
	}

	/**
	 * Инициализация запчасти
	 *
	 * Проверяет существование запчасти
	 *
	 * @param string $spare Запчасть
	 *
	 * @return string|bool Запчасть, если найдена, иначе false
	 */
	public static function spares(string $spare): string|bool
	{
		// Поиск запчастей и выход (успех)
		return match (mb_strtolower($spare)) {
			'цевьё' => 'Цевьё',
			default => false
		};
	}

	/**
	 * Главное меню
	 *
	 * Команда: /start
	 *
	 * @param Context $ctx
	 *
	 * @return void
	 */
	public static function menu(Context $ctx): void
	{
		// Инициализация клавиатуры
		$keyboard = [
			[
				['text' => '🛒 Каталог', 'web_app' => ['url' => 'https://arming.dev.mirzaev.sexy']]
			],
			[
				['text' => '🏛️ О компании'],
				['text' => '💬 Контакты']
			],
			[
				['text' => '🎯 Сообщество']
			]
		];

		if ($ctx->get('account')?->access['settings']) $keyboard[] = [['text' => '⚙️  Настройки']];

		// Отправка сообщения
		$ctx->sendMessage(
			static::unmarkdown(<<<TXT
		Это сообщение будет отображаться (оно должно быть обязательно) при вызове главного меню командой /start (создаёт кнопки меню снизу)
		TXT),
			[
				'reply_markup' => [
					'keyboard' => $keyboard,
					'resize_keyboard' => true
				],
				'disable_notification' => true
			]
		);
	}

	/**
	 * Начало работы с чат-роботом
	 *
	 * Команда: /start
	 *
	 * @param Context $ctx
	 *
	 * @return void
	 */
	public static function start(Context $ctx): void
	{
		// Главное меню
		static::menu($ctx);
	}

	/**
	 * Контакты
	 *
	 * Команда: /contacts
	 *
	 * @param Context $ctx
	 *
	 * @return void
	 */
	public static function contacts(Context $ctx): void
	{
		// Отправка сообщения
		$ctx->sendMessage(static::unmarkdown(<<<TXT
		Здесь придумать текст для раздела "Контакты"
		TXT), [
			'reply_markup' => [
				'inline_keyboard' => [
					[
						['text' => '⚡ Связь с менеджером', 'url' => 'https://t.me/iarming'],
					],
					[
						['text' => '📨 Почта', 'callback_data' => 'mail']
					],
					[
						['text' => '🪖 Сайт', 'url' => 'https://arming.ru'],
						['text' => '🛒 Wildberries', 'url' => 'https://arming.ru']
					]
				]
			],
			'link_preview_options' => [
				'is_disabled' => true
			],
			'disable_notification' => true
		]);
	}

	/**
	 * Почта
	 *
	 * @param Context $ctx
	 *
	 * @return void
	 */
	public static function _mail(Context $ctx): void
	{
		// Отправка сообщения
		$ctx->sendMessage(static::unmarkdown(<<<TXT
		[info@arming.ru](mailto::info@arming.ru)
		TXT, ['[', ']', '(', ')']), [
			'link_preview_options' => [
				'is_disabled' => true
			],
			'disable_notification' => true
		]);
	}

	/**
	 * Компания
	 *
	 * Команда: /company
	 *
	 * @param Context $ctx
	 *
	 * @return void
	 */
	public static function company(Context $ctx): void
	{
		// Отправка сообщения
		$ctx->sendMessage(
			static::unmarkdown(<<<TXT
		Здесь придумать текст для раздела "Компания"
		TXT),
			/* [
		'reply_markup' => [
			'inline_keyboard' => [
				[
					['text' => '⚡ Связь с менеджером', 'url' => 'https://git.mirzaev.sexy/mirzaev/mashtrash'],
					['text' => '📨 Почта', 'text' => ''],
				],
				[
					['text' => '🪖 Сайт', 'url' => '']
					['text' => '🛒 Wildberries', 'url' => '']
				]
			]
		],
		'link_preview_options' => [
			'is_disabled' => true
		]
	] */
		);
	}

	/**
	 * Сообщество
	 *
	 * Команда: /community
	 *
	 * @param Context $ctx
	 *
	 * @return void
	 */
	public static function community(Context $ctx): void
	{
		// Отправка сообщения
		$ctx->sendMessage(static::unmarkdown(<<<TXT
		Здесь придумать текст для раздела "Сообщество"
		TXT), [
			'reply_markup' => [
				'inline_keyboard' => [
					[
						['text' => '💬 Основной чат', 'url' => 'https://t.me/arming_zone'],
					]
				]
			],
			'link_preview_options' => [
				'is_disabled' => true
			],
			'disable_notification' => true
		]);
	}

	/**
	 * Настройки (доступ только авторизованным)
	 *
	 * Команда: /settings
	 *
	 * @param Context $ctx
	 *
	 * @return void
	 */
	public static function settings(Context $ctx): void
	{
		if ($ctx->get('account')?->access['settings']) {
			// Авторизован доступ к настройкам

			// Отправка сообщения
			$ctx->sendMessage(
				static::unmarkdown(<<<TXT
				Панель управления чат-роботом ARMING
				TXT),
				[
					'reply_markup' => [
						'inline_keyboard' => [
							[
								['text' => '📦 Импорт товаров', 'callback_data' => 'import_request'],
							]
						]
					],
					'link_preview_options' => [
						'is_disabled' => true
					],
					'disable_notification' => true
				]
			);
		} else {
			// Не авторизован доступ к настройкам

			// Отправка сообщения
			$ctx->sendMessage('⛔ *Нет доступа*');
		}
	}

	/**
	 * Запросить файл для импорта товаров (доступ только авторизованным)
	 *
	 * @param Context $ctx
	 *
	 * @return void
	 */
	public static function import_request(Context $ctx): void
	{
		if ($ctx->get('account')?->access['settings']) {
			// Авторизован доступ к настройкам

			// Отправка сообщения
			$ctx->sendMessage(static::unmarkdown('Отправьте документ в формате xlsx со списком товаров'))
				->then(function ($message) use ($ctx) {
					// Отправка файла
					$ctx->sendDocument(new InputFile(CATALOG_EXAMPLE), ['disable_notification' => true]);

					// Импорт файла
					$ctx->nextStep("import");
				});
		} else {
			// Не авторизован доступ к настройкам

			// Отправка сообщения
			$ctx->sendMessage('⛔ *Нет доступа*');
		}
	}

	/**
	 * Импорт товаров (доступ только авторизованным)
	 *
	 * @param Context $ctx
	 *
	 * @return void
	 */
	public static function import(Context $ctx): void
	{
		if ($ctx->get('account')?->access['settings']) {
			// Авторизован доступ к настройкам

			// Инициализация документа
			$document = $ctx->getMessage()?->getDocument();

			if ($document instanceof telegram_document) {
				// Инициализирован документ

				// Инициализация файла
				$ctx->getFile($document->getFileId())->then(function ($file) use ($ctx) {

					if ($file->getFileSize() <= 50000000) {
						// Не превышает 50 мегабайт (50000000 байт) размер файла

						if ($file->getFilePath()['extension'] === 'xlsx') {
							// Имеет расширение xlsx файл

							// Сохранение файла
							file_put_contents(STORAGE . DIRECTORY_SEPARATOR . 'import.xlsx', file_get_contents('https://api.telegram.org/file/bot' . KEY . '/' . $file->getFilePath()));

							// Инициализация счётчика загруженных товаров
							$loaded = $created = $updated = $deleted = $old = $new = 0;



							// Отправка сообщения
							$ctx->sendMessage(<<<TXT
							*Загружено для обработки:* $loaded

							*Добавлено:* $created
							*Обновлено:* $updated
							*Удалено:* $deleted

							*Опубликовано в магазине:* $old \-\> *$new*
							TXT)
								->then(function ($message) use ($ctx) {
									// Завершение диалога
									$ctx->endConversation();
								});
						} else {
							// Не имеет расширение xlsx файл

							// Отправка сообщения
							$ctx->sendMessage(static::unmarkdown('Файл должен иметь расширение xlsx'));
						}
					} else {
						// Превышает 50 мегабайт (50000000 байт) размер файла

						// Отправка сообщения
						$ctx->sendMessage(static::unmarkdown('Размер файла не должен превышать 50 мегабайт'));
					}
				});
			} else {
				// Не инициализирован документ

				// Отправка сообщения
				$ctx->sendMessage(static::unmarkdown('Отправьте документ в формате xlsx со списком товаров'));
			}
		} else {
			// Не авторизован доступ к настройкам

			// Отправка сообщения
			$ctx->sendMessage('⛔ *Нет доступа*');
		}
	}

	/**
	 * Инициализация аккаунта (middleware)
	 *
	 * @param Context $ctx
	 * @param Node $next
	 *
	 * @return void
	 */
	public static function account(Context $ctx, Node $next): void
	{
		// Выполнение заблокировано?
		if ($ctx->get('stop')) return;

		// Инициализация аккаунта Telegram
		$telegram = $ctx->getEffectiveUser();

		// Инициализация аккаунта
		$account = account::initialization($telegram->getId(), $telegram);

		if ($account) {
			// Инициализирован аккаунт

			if ($account->banned) {
				// Заблокирован аккаунт

				// Отправка сообщения
				$ctx->sendMessage('⛔ *Ты заблокирован*')
					->then(function ($message) use ($ctx) {
						// Завершение диалога
						$ctx->endConversation();
					});

				// Блокировка дальнейшего выполнения
				$ctx->set('stop', true);
			} else {
				// Не заблокирован аккаунт

				// Запись в буфер
				$ctx->set('account', $account);

				// Продолжение выполнения
				$next($ctx);
			}
		} else {
			// Не инициализирован аккаунт
		}
	}

	/**
	 * Инициализация статуса технических работ (middleware)
	 *
	 * @param Context $ctx
	 * @param Node $next
	 *
	 * @return void
	 */
	public static function suspension(Context $ctx, Node $next): void
	{
		// Выполнение заблокировано?
		if ($ctx->get('stop')) return;

		// Поиск технических работ
		$suspension = suspension::search();

		if ($suspension && $suspension->targets['chat-robot']) {
			// Найдена активная приостановка

			// Инициализация аккаунта
			$account = $ctx->get('account');

			if ($account) {
				// Инициализирован аккаунт

				foreach ($suspension->access as $type => $status) {
					// Перебор статусов доступа

					if ($status && $account->{$type}) {
						// Авторизован аккаунт

						// Продолжение выполнения
						$next($ctx);

						// Выход (успех)
						return;
					}
				}
			}

			// Инициализация сообщения
			$message = "⚠️ *Работа приостановлена*\n*Оставшееся время\:* " . $suspension->message($account->language ?? controller::$settings?->language);

			// Добавление описания причины приостановки, если найдена
			if (!empty($suspension->description)) $message .= "\n\n" . $suspension->description[$account->language ?? controller::$settings?->language] ?? array_values($suspension->description)[0];

			// Отправка сообщения
			$ctx->sendMessage($message)
				->then(function ($message) use ($ctx) {
					// Завершение диалога
					$ctx->endConversation();
				});

			// Блокировка дальнейшего выполнения
			$ctx->set('stop', true);
		} else {
			// Не найдена активная приостановка

			// Продолжение выполнения
			$next($ctx);
		}
	}

	/**
	 * Write
	 *
	 * Write a property into an instance of the ArangoDB document
	 *
	 * @param string $name Name of the property
	 * @param mixed $value Content of the property
	 *
	 * @return void
	 */
	public function __set(string $name, mixed $value = null): void
	{
		// Write to the property into an instance of the ArangoDB document and exit (success)
		$this->document->{$name} = $value;
	}

	/**
	 * Read
	 *
	 * Read a property from an instance of the ArangoDB docuemnt
	 * 
	 * @param string $name Name of the property
	 *
	 * @return mixed Content of the property
	 */
	public function __get(string $name): mixed
	{
		// Read a property from an instance of the ArangoDB document and exit (success)
		return match ($name) {
			'arangodb' => $this::$arangodb,
			default => $this->document->{$name}
		};
	}

	/**
	 * Delete
	 *
	 * Deinitialize the property in an instance of the ArangoDB document
	 *
	 * @param string $name Name of the property
	 *
	 * @return void
	 */
	public function __unset(string $name): void
	{
		// Delete the property in an instance of the ArangoDB document and exit (success)
		unset($this->document->{$name});
	}

	/**
	 * Check of initialization
	 *
	 * Check of initialization of the property into an instance of the ArangoDB document
	 *
	 * @param string $name Name of the property
	 *
	 * @return bool The property is initialized?
	 */
	public function __isset(string $name): bool
	{
		// Check of initializatio nof the property and exit (success)
		return isset($this->document->{$name});
	}

	/**
	 * Execute a method
	 *
	 * Execute a method from an instance of the ArangoDB document
	 *
	 * @param string $name Name of the method
	 * @param array $arguments Arguments for the method
	 *
	 * @return mixed Result of execution of the method
	 */
	public function __call(string $name, array $arguments = []): mixed
	{
		// Execute the method and exit (success)
		if (method_exists($this->document, $name)) return $this->document->{$name}($arguments);
	}
}
