<?php

declare(strict_types=1);

namespace mirzaev\arming_bot\models;

// Files of the project
use mirzaev\arming_bot\models\core,
	mirzaev\arming_bot\controllers\core as controller,
	mirzaev\arming_bot\models\settings,
	mirzaev\arming_bot\models\traits\document as arangodb_document_trait,
	mirzaev\arming_bot\models\interfaces\document as arangodb_document_interface;

// Framework for ArangoDB
use mirzaev\arangodb\collection,
	mirzaev\arangodb\document;

// Library для ArangoDB
use ArangoDBClient\Document as _document;

// Built-in libraries
use exception,
	datetime;

/**
 * Model of a suspension
 *
 * @package mirzaev\arming_bot\models
 * @author Arsen Mirzaev Tatyano-Muradovich <arsen@mirzaev.sexy>
 */
final class suspension extends core implements arangodb_document_interface
{
	use arangodb_document_trait;

	/**
	 * Name of the collection in ArangoDB
	 */
	final public const string COLLECTION = 'suspension';

	/**
	 * Search for active suspension
	 *
	 * @param array &$errors Registry of errors
	 *
	 * @return static|null Object implements an instance of suspension from ArangoDB
	 */
	public static function search(array &$errors = []): static|null
	{
		try {
			if (collection::init(core::$arangodb->session, self::COLLECTION)) {
				if ($document = collection::search(core::$arangodb->session, sprintf("FOR d IN %s FILTER d.end > %u SORT d.end DESC LIMIT 1 RETURN d", self::COLLECTION, time()))) {
					// Найдены активные настройки

					// Инициализация объекта настроек
					$suspension = new static;

					// Запись инстанции документа в объект
					$suspension->document = $document;

					// Возврат (успех)
					return $suspension;
				} else return null;
			} else throw new exception('Failed to initialize document collection: ' . self::COLLECTION);
		} catch (exception $e) {
			// Write to the registry of errors
			$errors[] = [
				'text' => $e->getMessage(),
				'file' => $e->getFile(),
				'line' => $e->getLine(),
				'stack' => $e->getTrace()
			];
		}

		// Exit (fail)
		return null;
	}

	/**
	 * Generate message about remaining time
	 *
	 * @param string|null $language Language of the generated text (otherwise used from settings.language)
	 * @param array &$errors Registry of errors
	 *
	 * @return string|null Text: "? days, ? hours and ? minutes"
	 */
	public function message(?string $language = null, array &$errors = []): ?string
	{
		try {
			// Initializing default value
			$language ??= controller::$settings?->language ?? 'en';

			// Initializing the time until the suspension ends
			$difference = date_diff(new datetime('@' . $this->document->end), new datetime());

			// Generate text about remaining time and exit (success)
			return sprintf(
				'%u %s, %u %s и %u %s',
				$difference->d,
				match ($difference->d > 20 ? $difference->d % 10 : $difference->d % 100) {
					1 => match ($language) {
						'ru' => 'день',
						'en' => 'day',
						default => 'day'
					},
					2, 3, 4 => match ($language) {
						'ru' => 'дня',
						'en' => 'days',
						default => 'days'
					},
					default => match ($language) {
						'ru' => 'дней',
						'en' => 'days',
						default => 'days'
					}
				},
				$difference->h,
				match ($difference->h > 20 ? $difference->h % 10 : $difference->h % 100) {
					1 => match ($language) {
						'ru' => 'час',
						'en' => 'hours',
						default => 'hour'
					},
					2, 3, 4 => match ($language) {
						'ru' => 'часа',
						'en' => 'hours',
						default => 'hours'
					},
					default => match ($language) {
						'ru' => 'часов',
						'en' => 'hours',
						default => 'hours'
					}
				},
				$difference->i,
				match ($difference->i > 20 ? $difference->i % 10 : $difference->i % 100) {
					1 => match ($language) {
						'ru' => 'минута',
						'en' => 'minute',
						default => 'minute'
					},
					2, 3, 4 => match ($language) {
						'ru' => 'минуты',
						'en' => 'minutes',
						default => 'minutes'
					},
					default => match ($language) {
						'ru' => 'минут',
						'en' => 'minutes',
						default => 'minutes'
					}
				}
			);
		} catch (exception $e) {
			// Write to the registry of errors
			$errors[] = [
				'text' => $e->getMessage(),
				'file' => $e->getFile(),
				'line' => $e->getLine(),
				'stack' => $e->getTrace()
			];
		}

		// Exit (fail)
		return null;
	}
}
