<?php

declare(strict_types=1);

namespace mirzaev\arming_bot\models\traits;

// Files of the project
use mirzaev\arming_bot\models\core;

// Library для ArangoDB
use ArangoDBClient\Document as _document;

/**
 * Trait for implementing a document instance from ArangoDB
 *
 * @package mirzaev\arming_bot\models\traits
 * @author Arsen Mirzaev Tatyano-Muradovich <arsen@mirzaev.sexy>
 */
trait document
{
	/**
	 * An instance of the ArangoDB document from ArangoDB
	 */
	protected readonly _document $document;

	/**
	 * Write or read document
	 *
	 * @param _document|null $document Instance of document from ArangoDB
	 *
	 * @return _document|null Instance of document from ArangoDB
	 */
	public function __document(?_document $document): ?_document
	{
		// Write a property storing a document instance to ArangoDB
		if ($document) $this->document = $document;

		// Read a property storing a document instance to ArangoDB and exit (success)
		return $this->document;
	}

	/**
	 * Write
	 *
	 * Write a property into an instance of the ArangoDB document
	 *
	 * @param string $name Name of the property
	 * @param mixed $value Content of the property
	 *
	 * @return void
	 */
	public function __set(string $name, mixed $value = null): void
	{
		// Write to the property into an instance of the ArangoDB document and exit (success)
		$this->document->{$name} = $value;
	}

	/**
	 * Read
	 *
	 * Read a property from an instance of the ArangoDB docuemnt
	 * 
	 * @param string $name Name of the property
	 *
	 * @return mixed Content of the property
	 */
	public function __get(string $name): mixed
	{
		// Read a property from an instance of the ArangoDB document and exit (success)
		return match ($name) {
			'arangodb' => core::$arangodb,
			default => $this->document->{$name}
		};
	}

	/**
	 * Delete
	 *
	 * Deinitialize the property in an instance of the ArangoDB document
	 *
	 * @param string $name Name of the property
	 *
	 * @return void
	 */
	public function __unset(string $name): void
	{
		// Delete the property in an instance of the ArangoDB document and exit (success)
		unset($this->document->{$name});
	}

	/**
	 * Check of initialization
	 *
	 * Check of initialization of the property into an instance of the ArangoDB document
	 *
	 * @param string $name Name of the property
	 *
	 * @return bool The property is initialized?
	 */
	public function __isset(string $name): bool
	{
		// Check of initializatio nof the property and exit (success)
		return isset($this->document->{$name});
	}

	/**
	 * Execute a method
	 *
	 * Execute a method from an instance of the ArangoDB document
	 *
	 * @param string $name Name of the method
	 * @param array $arguments Arguments for the method
	 *
	 * @return mixed Result of execution of the method
	 */
	public function __call(string $name, array $arguments = []): mixed
	{
		// Execute the method and exit (success)
		return method_exists($this->document, $name) ?$this->document->{$name}($arguments) ?? null : null;
	}
}
